/*========================================================================*\

Copyright (c) 1999  Nicolai Langfeldt

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL NICOLAI LANGFELDT, PAUL VOJTA, OR ANYONE ELSE BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Names:
  janl: Nicolai Langfeldt (janl@math.uio.no)
  rmz: Rainer Menzner (Rainer.Menzner@neuroinformatik.ruhr-uni-bochum.de)

History:
  18/02/1999, janl: Initial version, first public alpha release of xdvi-t1.
  22/02/1999, rmz: Changes to make scaling and antialiasing work right.
  24/02/1999, janl: Read afm files to get alignments right.  type1 char
  	widhts are to coarse.

Bugs:
  - Every kpse_find_file call presents a memory leak.  The pathname
    buffer is not freed.  Does it need to or does kpse_find_file use
    a static buffer?

\*========================================================================*/

#include "xdvi-config.h"

#ifdef T1

#include <assert.h>

#if defined(HAVE_STRING_H)
#include <string.h>
#else
extern char *strtok P2H(char *, const char *)
#endif

#include "dvi.h"
#include <kpathsea/tex-file.h>

#include <t1lib.h>
#include <t1libx.h>

/* General documentation:

   encodings[n]: Loaded from map file
   [0] = 8r,	<vector>
   [1] = 8c,	<vector>

   The vectors are demand loaded in 'load_encoded_font'

   fontmaps[n]: Loaded from map file
   idx	 texname	file	enc	ext	sl
   [0] = pcrr8rn,	pcrr,	0,	850,	0
   [1] = putro8r,	putr,	0,	0,	1670

   0 extention and slant means no extention/slant.  If the input
   values are decimal (less than 10) they're multiplied by 1000 and
   10000 respectively to obtain fixed-point integer values.  Integer
   values have the advandage of being testable for exact match (== 0)
   on all architectures (it's true!  Not all architectures can do
   (reliable) exact matching of floatingpoint values)

   texfonts[n]: Built as xdvi loads fonts
   idx	 texname		file			t1id	width
   [0] = cmr10 (unencoded),	/path/cmr10.pfb,	0 	0
   [1] = ptmr (unencoded),	/path/prmr.pfa,		1	-1
   [2] = ptmr8r,		NULL,			2	1

   Fonts whose raw encoding/slanting/extention match those tex want
   can just be loaded.  For fonts what have a map entry, or which have
   an implied encoding (ptmr8r => 8r implied coding) we must first
   load the raw/unmodified font (it not yet loaded).  Then use
   T1_CopyFont and modify the encoding/slanting/extention of the copy.
   -1 in the width index field means that the widht data is not
   loaded.  -2 means it's not available.

   Fonts that are copies of other fonts have NULL filenames.  If any
   font files are to be reused based on the file name the raw font
   must be reused, not a copy.

   widthinfo[n]: Demand loaded as the fonts are used.
   idx	 texname	widths
   [0] = cmr10		...
   [1] = ptmr8r		...

   The widthinfo is loaded from the font tfm file since the type1
   width information isn't precise enough to avoid accumulating
   rounding errors that make things visibly unaligned.  This is
   esp. noticeable in tables where the vline column separators become
   unaligned.

*/

typedef struct encoding {
  char *enc;		/* This may be NULL.  From dvips maps we get 
			   encodings loaded by filename */
  char *file;
  char **vector;
} encoding;

encoding *encodings;	/* Dynamic array */
#define ENCSTART 16	/* Start size.  Normaly there will not be many. */
#define ENCGROW 16	/* Grow by when full */
int maxenc=ENCSTART-1;	/* Current size of array */
int enclidx=-1;		/* Current last index in array */

typedef struct fontmap {
  char *texname;
  char *filename;
  int  enc;	/* Index in encoding array */
  int  extention; /* Fixed point, *1000 */
  int  slant;	/* Fixed point, *10000, some font slantings have 4 
		   significant digits, all after the decimalpoint */
} fontmap;

fontmap *fontmaps;
#define FNTMAPSTART 256	/* This should keep most of them happy */
#define FNTMAPGROW 64
int maxmap=FNTMAPSTART-1;
int maplidx=-1;

typedef struct texfont {
  char *texname;
  char *filename;	/* If this is NULL the font is a copy */
  int  t1id;
  int  tfmidx;		/* Index in tfminfo array */
} texfont;


texfont *texfonts;
#define TEXFNTSTART 64
#define TEXFNTGROW 32
int maxtex=TEXFNTSTART-1;
int texlidx=-1;

typedef struct tfminfos {
  char *texname;
  long designsize;
  long widths[256];
} tfminfos;

tfminfos *tfminfo;
#define TEXWSTART 64
#define TEXWGROW 32
int maxw=TEXWSTART-1;
int wlidx=-1;

/* For file reading */
#define BUFFER_SIZE 512
char buffer[BUFFER_SIZE];

/* This is the conversion factor from TeXs pt to TeXs bp (big point).
   Maximum machine precision is needed to keep it acurate. */
double bp_fac=72.0/72.27;

/* Try to convert from PS charspace units to DVI units with minimal
   loss of significant digits */
#define t1_dvi_conv(x)	((((int) (x)) << 16)/1000)
/* Convert from DVI units to pt */
#define dvi_pt_conv(x)	(((int) (x)) >> 16)

/* Convert from TFM units to DVI units */
#define tfm_dvi_conv(x) (((int) (x)) >> 1)

/* ********************** FONT AND ENCODING LOADING ******************** */

char **load_vector P1C(char *,enc)

{
  char *filename;
  char **vector=NULL;

  filename=kpse_find_file(enc,kpse_program_text_format,0);
  
  if (filename!=NULL) {
    vector=T1_LoadEncoding(filename);
    if (vector==NULL) {
      fprintf(stderr,"xdvi: Unable to load encoding vector %s from %s\n",
	      enc,filename);
      perror("xdvi");
      exit(1);
    }
    fprintf(stderr,"Encoding %s loaded from %s\n",enc,filename);
  } else
    fprintf(stderr,"Unable to load encoding vector %s\n",enc);

  return vector;
}


int add_tfm P1C(char *,texname)

{

  /* Load font metrics if not already loaded.  Return index into
     fontmetrics array, or -1 if it failed */

  int i;
  char *filename;

  for (i=0;i<=wlidx;i++) 
    if (strcmp(tfminfo[i].texname,texname)==0)
      return i;

  if (wlidx==maxw) {
    maxw += TEXWGROW;
    tfminfo=realloc(tfminfo,sizeof(*tfminfo)*maxw);
    if (tfminfo==NULL) {
      fprintf(stderr,"xdvi: Error: Out of memory.\n");
      exit(1);
    }
  }

  wlidx++;

  tfminfo[wlidx].texname=texname;

  if (!tfmload(texname,&tfminfo[wlidx].designsize,tfminfo[wlidx].widths)) {
      fprintf(stderr,"xdvi: Could not find/load tfm file for %s. "
	      "xdvi: Expect ugly output.\n");

    wlidx--;
    return -1;
  }

  return wlidx;
}


int add_texfont P3C(char *,texname, char *,filename, int, id)

{
  /* Add font to list.  The caller has already checked that the font
     is not already on the list.  Return index at which it was
     inserted. */

  if (texlidx==maxtex) {
    maxtex += TEXFNTGROW;
    texfonts=realloc(texfonts,sizeof(*texfonts)*maxtex);
    if (encodings==NULL) {
      fprintf(stderr,"xdvi: Error: Out of memory\n");
      exit(1);
    }
    fprintf(stderr,"Enlarged font table from %d to %d entries\n",
	    texlidx, maxtex);
  }

  texlidx++;

  /* Check that our assumption of texfont index == t1id holds.
     It might not in future versions of t1lib. */
  assert(texlidx==id);

  texfonts[texlidx].texname=strdup(texname);
  texfonts[texlidx].filename=(filename==NULL)?NULL:strdup(filename);
  texfonts[texlidx].t1id=id;
  texfonts[texlidx].tfmidx=-1;

  return texlidx;
}


int load_encoded_font P6C(char *,name, char *,alias, char *,filename,
			  int, enc, int, ext, int ,sl)


{
  /* Load the font, encoding, extending and slanting it as
     prescribed.  Return the texfont array index.

     The 'file' param should be the font-file-stub-name, such as
     ptmr8a or cmr10, this should then be used for a kpse_find_file
     call to locate the actual file.

     There are two main scenarios for this procedure:

     1: params = ptmr8r, ptmr8a, NULL, enc-idx-of-8r, 0, 0

     This should result in two texfonts entries:
	
       ptmr8a, ...texmf...ptmr8a.pfa, t1id-of-ptmr8a, -1
       ptmr8r, (null),                t1id-of-ptmr8r-which-is-copy-of-8a, -1

     IFF ptmr8a is already loaded a copy should be made and only the
     second entry should be generated on this call.

     2: params = cmr10, NULL, NULL, -1, 0, 0

     should lead to one new entry:

       cmr10, /.../texmf/.../cmr10.pfb, t1id-of-cmr10, -1

 */

  int id;
  int idx=-1;
  int i;
  int added=0;
  int isalias=0;
  char *canon;
  
  /* Already loaded/constructed by that name? */
  for (i=0;i<texlidx;i++) {
    if (strcmp(texfonts[i].texname,name)==0) {
      fprintf(stderr,"Type1 font already loaded: %s\n",name);
      return i;
    }
  }

  if (alias!=NULL) {
    /* Is the 'raw' name (alias) known? */
    for (i=0;i<texlidx;i++) {
      if (alias!=NULL && strcmp(texfonts[i].texname,alias)==0) {
	fprintf(stderr,"Found alias of %s: %s\n",name,alias);
	idx=i;
	isalias=1; 
	/* Must make copy of t1 font to set up the new one... if there
           are setup params to modify it */
	break;
      }
    }
  }

  
  /* Not loaded by any name.  Find and load */
  if (idx==-1) {
    canon=(alias==NULL)?(name):(alias);

    if (filename==NULL) {
      /* Full filename was not passed to us.  Find it */
      filename=kpse_find_file(canon,kpse_type1_format,0);
      if (filename==NULL) return -1;
    }

    id=T1_AddFont(filename);

    fprintf(stderr,"%s loaded as t1lib font %d\n",
	    filename, id);

    /* Set the afm filename.  Always before loading. */
    filename=kpse_find_file(canon,kpse_afm_format,0);
    if (filename!=NULL) 
      T1_SetAfmFileName(id,filename);

    /* Always load before encoding */
    if (id<0 || T1_LoadFont(id)==-1) {
      fprintf(stderr,"Could not load font %s\n",filename);
      return -1;
    }

    if (alias==NULL)
      /* Not alias? Then we're just about finished */
      return add_texfont(name,filename,id);

    idx=add_texfont(alias,filename,id);
  }

  /* If there is nothing further to do, just return index */
  if (enc==-1 && ext==0 && sl==0) return add_texfont(name,NULL,id);

  /* Copy the raw font so we can muck with it */
  id = T1_CopyFont(texfonts[idx].t1id);

  idx = add_texfont(name,NULL,id);

  if (enc!=-1) {
    /* Demand load vector */
    if (encodings[enc].vector==NULL)
      encodings[enc].vector=load_vector(encodings[enc].file);

    if (encodings[enc].vector==NULL) {
      fprintf(stderr,"xdvik: Could not load encoding file %s for vector %s\n",
	      encodings[enc].file,encodings[enc].enc);
    } else {
      if (T1_ReencodeFont(id,encodings[enc].vector)!=0) {
	fprintf(stderr,"Re-encoding %s did not work\n",name);
      }
    }
  }

  if (ext)
    T1_ExtendFont(id,ext/1000.0);

  if (sl) 
    T1_SlantFont(id,sl/10000.0);
  
  return idx;
}



int find_T1_font P1C(char *,name)

     /* Find the font with the given name and return the t1lib id */

{
  int i,id;	/* Iterator, t1 font id */
  char *filename;
  int fl,el;	/* Fontname length, encoding name length */
  int encoded=-1;
  char *mname;	/* Modified font name */

  encoded=0;

  /* Zeroth: Check the already loaded fonts */
  for (i=0;i<=texlidx;i++) {
    if (strcmp(name,texfonts[i].texname)==0)
      return i;
  }

  /* First: the mapings */

  for (i=0;i<=maplidx;i++) {
    if (strcmp(name,fontmaps[i].texname)==0) {
      /* We have a winnner! */
      return load_encoded_font(fontmaps[i].texname, fontmaps[i].filename,
			       NULL, fontmaps[i].enc, fontmaps[i].extention,
			       fontmaps[i].slant);
    }
  }
  
  /* Second: the bare name */

  filename=kpse_find_file(name,kpse_type1_format,0);

  if (filename!=NULL)
    return load_encoded_font(name, NULL, filename, -1, 0, 0);

  /* Third: Implied encoding? */
  fl=strlen(name);

  for (i=0; i<=enclidx; i++) {
    if (encodings[i].enc==NULL)
      continue;
    el=strlen(encodings[i].enc);
    
    if (strcmp(name+(fl-el),encodings[i].enc)==0) {
      /* Found a matching encoding */
      fprintf(stderr,"Encoding match: %s - %s\n",name,encodings[i].enc);
      mname=malloc(fl+2);
      strcpy(mname,name);
      mname[fl-el]='\0';
      /* If we had 'ptmr8r' the we now look for 'ptmr' */
      filename=kpse_find_file(mname, kpse_type1_format,0);
      if (filename==NULL) {
	/* No? Look for ptmr8a. 8a postfix is oft used on raw fonts */
	strcat(mname,"8a");  /* 'adobe' encoding. */
	filename=kpse_find_file(mname,kpse_type1_format,0);
      }
      if (filename!=NULL)
	return load_encoded_font(name, mname, filename, i, 0, 0);

      free(mname);
      /* Now we tried everything.  Time to give up. */
      return -1;

    } /* If (strcmp...) */
  } /* for */

  return -1;
}

/* ************************* CONFIG FILE READING ************************ */


int new_encoding P2C(char *,enc, char *,file)

{
  /* (Possebly) new encoding entered from .cfg file, or from dvips
     map.  When entered from dvips map the enc is null/anonymous */

  int i;

  /* Shirley!  You're jesting! */
  if (file==NULL) return -1;

  if (enc==NULL) {
    /* Enter by file name.  First check if the file has been entered
       already. */
    for (i=0; i<=enclidx; i++) {
      if (strcmp(encodings[i].file,file)==0)
	return i;
    }
  } else {
    /* Enter by encoding name.  Check if already loaded first. */
    for (i=0; i<=enclidx; i++) {
      if (encodings[i].enc!=NULL && strcmp(encodings[i].enc,enc)==0)
	return i;
    }
  }
  
  /* Bonafide new encoding */

  /* Any space left? */
  if (enclidx==maxenc) {
    maxenc += ENCGROW;
    encodings=realloc(encodings,sizeof(*encodings)*maxenc);
    if (encodings==NULL) {
      fprintf(stderr,"xdvi: Error: No memory for encoding map.?\n");
      exit(1);
    }
    fprintf(stderr,"Enlarged encoding map from %d to %d entries\n",
	    enclidx,maxenc);
  }

  enclidx++;
  fprintf(stderr,"New encoding #%d: '%s' -> '%s'\n",enclidx,enc,file);
  encodings[enclidx].enc=strdup(enc);
  encodings[enclidx].file=strdup(file);
  encodings[enclidx].vector=NULL;	/* Demand load later */

  if (encodings[enclidx].enc==NULL || encodings[enclidx].file==NULL) {
    fprintf(stderr,"xdvi: Ran out memory\n");
    exit(0);
  }
	  
  return enclidx;
}


void add_mapentry P4C(char *,name, char *,file, char *,vec, char *,spec)

{

  /* This is called from getpsinfo, once for each fontmap line.  We
   * will dutifully enter the information into our fontmap table */

  static char delim[]="\t ";
  char *last, *current;
  float number;
  int extend=0;
  int slant=0;
  int i;

#if 0
  fprintf(stderr,"Map entry: %s -> %s(%s) Spec: %s\n",name,file,vec,spec);
#endif

  if (spec!=NULL) {
    /* Try to analyze the postscript string.  We recognize two things:
       "n ExtendFont" and "m SlantFont".  n can be a decimal number in
       which case it's an extention factor, or a integer, in which
       case it's a charspace unit? In any case 850 = .85 */

    last=strtok(spec,delim);
    current=strtok(NULL,delim);
    while (current!=NULL) {
      if (strcmp(current,"ExtendFont")==0) {
	sscanf(last,"%f",&number);
	if (number<10.0)
	  extend = number * 1000.0;
	else 
	  extend = number;
#if 0
	fprintf(stderr,"EXTEND: %d\n",extend);
#endif
      } else if (strcmp(current,"SlantFont")==0) {
	sscanf(last,"%f",&number);
	slant = number * 10000.0;
#if 0
	fprintf(stderr,"SLANT: %d\n",slant);
#endif
      }
      last=current;
      current=strtok(NULL,delim);
    }
  }

  /* Any free space left? */
  if (maplidx==maxmap) {
    maxmap += FNTMAPGROW;
    /* Hope realloc isn't destructive anywhere :-) */
    fontmaps=realloc(fontmaps,sizeof(*fontmaps)*maxmap);
    if (fontmaps==NULL) {
      fprintf(stderr,"xdvi: Error: No memory for fontmap.  Buy some RAM?\n");
      exit(1);
    }
    fprintf(stderr,"Enlarged the fontmap from %d to %d entries\n",
	    maplidx,maxmap);
  }

  maplidx++;
  fontmaps[maplidx].texname=strdup(name);
  fontmaps[maplidx].filename=strdup(file);
  fontmaps[maplidx].enc=new_encoding(NULL,vec);
  fontmaps[maplidx].extention=extend;
  fontmaps[maplidx].slant=slant;

  if (fontmaps[maplidx].texname==NULL || fontmaps[maplidx].filename==NULL) {
    fprintf(stderr, "xdvi: Error: Dang, can't find any free memory!\n");
    exit(1);
  }
}


void read_cfg_file P1C(char *,file)

     /* Read config file */

{
  char *filename;
  FILE *mapfile;
  int len;
  char *keyword;
  char *enc;
  char *f;
  static const char delim[]="\t \n\r";

  if (file==NULL)
    file="xdvi.cfg";

  filename=kpse_find_file(file,kpse_program_text_format,1);
  if (filename==NULL) {
    fprintf(stderr,"xdvi: Warning: unable to find %s\n",file);
    return;
  }

  mapfile = fopen(filename,"r");

  if (mapfile==NULL) {
    fprintf(stderr,"xdvi: Error: Unable to open %s for reading: ",filename);
    perror("");
    exit(1);
  }

  fprintf(stderr,"Reading cfg file %s\n",filename);

  while (fgets(buffer, BUFFER_SIZE, mapfile)!=NULL) {
    len=strlen(buffer);
    if (buffer[len-1]!='\n') {
      fprintf(stderr, "xdvi: Error: Line longer than %d in %s. %d\n",
	      BUFFER_SIZE,filename,(int)buffer[len]);
      exit(1);
    }

    keyword=buffer;

    while (keyword[0]!='\0' && (keyword[0]==' ' || keyword[0]=='\t')) 
      keyword++;

    /* % in first column is a correct comment */
    if (keyword[0]=='%' || keyword[0]=='\0' || keyword[0]=='\n') 
      continue;

    keyword=strtok(keyword,delim);

    if (strcmp(keyword,"encoding")==0) {
      enc=strtok(NULL,delim);
      f=strtok(NULL,delim);
#if 0
      fprintf(stderr,"ENCODING: '%s' -> '%s'\n",enc,f);
#endif
      new_encoding(enc,f);
    } else if (strcmp(keyword,"dvipsmap")==0) {
      f=strtok(NULL,delim);
#if 0
      fprintf(stderr,"DVIPSMAP: '%s'\n",f);
#endif
      if (!getpsinfo(f)) {
	fprintf(stderr,"Could not load dvips map %s\n",f);
	exit(1);
      }
    } else {
      fprintf(stderr,"xdvi: Error: Unknown keyword: '%s' in %s",
	      keyword,filename);
      exit(1);
    }
  }

  fclose(mapfile);

  return;
}


/* **************************** GLYPH DRAWING *************************** */


/* Set character# ch */

#ifdef TEXXET
void set_t1_char P2C(wide_ubyte, cmd, wide_ubyte, ch)
#else
long set_t1_char P1C(wide_ubyte, ch)
#endif

{
  GC gc;
  int xpos = PXL_H - currwin.base_x ;
  /* + 1 here, seems t1lib and xdvi rounds the baseline differently */
  int ypos = PXL_V - currwin.base_y + 1;

  float size=currinf.fontp->fsize * bp_fac / pixels_per_inch * 
    dvi_pt_conv(currinf.fontp->scale);

  int id=currinf.fontp->t1id;
  int dvi_h_sav;

  long gw ;

  /* Is tfm width data loaded yet? */
  if (texfonts[id].tfmidx==-1)
    texfonts[id].tfmidx=add_tfm(texfonts[id].texname);

  if (texfonts[id].tfmidx==-2)
    /* Fallback when tfm data is missing.  Not precise enough for
       tables with vertical lines. */
    gw = t1_dvi_conv(T1_GetCharWidth(currinf.fontp->t1id, ch)*size)*
    pixels_per_inch/72;
  else
    /* Use the widht from the tfm file */
    gw = tfminfo[texfonts[id].tfmidx].widths[ch] * currinf.fontp->dimconv;

#ifdef HTEX
  gc = HTeXreflevel > 0? highGC: foreGC;
#else
  gc = foreGC;
#endif

#ifdef TEXXET
  dvi_h_sav = DVI_H;
  if (currinf.dir < 0) DVI_H -= gw;

  /* FIXME: There is something about scan_frame in set_char, I have no
     idea what it does, and dvi-draw does not export it either. */
#endif

#ifdef GREY  
  if (use_grey && !(shrink_factor==1)) {
    T1_AASetCharX(currwin.win, gc, T1_OPAQUE, xpos, ypos, 
		  currinf.fontp->t1id, ch, size/shrink_factor, NULL);
  } else
#endif
    {
      T1_SetCharX(currwin.win, gc, T1_OPAQUE, xpos, ypos, currinf.fontp->t1id,
		  ch, size/shrink_factor, NULL);
    }

#ifndef TEXXET
  return gw;
#else
# ifdef Omega
  if ((cmd == PUT1) || (cmd == PUT2))
# else
  if (cmd == PUT1)
# endif
    DVI_H = dvi_h_sav;
  else
    if (currinf.dir > 0) 
      DVI_H += gw;
#endif
}


void read_T1_char P2C(struct font *,fontp, ubyte, ch)

{
  /* Should never be called */
  fprintf(stderr,"BUG: asked to load %c\n",ch);
}


void init_t1()

{
  /* Initialize t1lib for drawing marks on screen */

  if ((T1_InitLib(LOGFILE|IGNORE_CONFIGFILE|IGNORE_FONTDATABASE)==NULL)) {
    fprintf(stderr, "xdvi: Error: Initialization of t1lib failed\n");
    exit(1);
  }
  T1_SetLogLevel(T1LOG_DEBUG);

  T1_SetDeviceResolutions((float)pixels_per_inch,(float)pixels_per_inch);

  T1_SetX11Params(DISP, our_visual, our_depth, our_colormap);
  /* T1_AA_LOW seems better than T1_AA_HIGH */
  T1_AASetLevel(T1_AA_LOW);
  T1_AASetBitsPerPixel(our_depth);

  encodings=malloc(sizeof(*encodings)*ENCSTART);
  fontmaps=malloc(sizeof(*fontmaps)*FNTMAPSTART);
  texfonts=malloc(sizeof(*texfonts)*TEXFNTSTART);
  tfminfo=malloc(sizeof(*tfminfo)*TEXWSTART);

  if (encodings==NULL || fontmaps==NULL || texfonts==NULL) {
    fprintf(stderr,"xdvi: Error: No memory for datastructures.\n");
    exit(1);
  }

  /* Malloc mem before reading config files! */
  read_cfg_file(NULL);
}

#endif /* t1 */
